/*	Chapter 3. RecordAccess. */
/*	Demonstrates simple record management in a file.
	The file is specified as a collection of an arbitrary
	number of records (The number is specified on the command
	line). Each record has a character string (limited length),
	times of creation and last update, and an update count.
	The file has an 8-byte HEADER containing the totol number
	of record slots and the total number of non-empty records.
	The user can interactively create, update, and modify
	records.
	Usage: RecordAccess FileName [nrec]
	If nrec is omitted, FileName must already exist. 
	If nrec is supplied, FileName is recreated (destroying any existing file)
	If the number of records is large, a sparse file is recommended.
	The techniques here could also be used if a hash function
	determined the record location (this would be a simple
	prorgam enhancment). */

/* This program illustrates:
	1. Random file access.
	2. LARGE_INTEGER arithmetic and using the 64-bit file positions. 
	3. Record update in place.
	4. File initialization to 0 (this will not work under Windows 9x
		or with a FAT file system. 
*/

#include "EvryThng.h"
#include <WINIOCTL.H>

#define STRING_SIZE 256
typedef struct _RECORD { /* File record structure */
	DWORD		ReferenceCount;  /* 0 meands an empty record */
	SYSTEMTIME	RecordCreationTime;
	SYSTEMTIME  RecordLastRefernceTime;
	SYSTEMTIME	RecordUpdateTime;
	TCHAR		DataString[STRING_SIZE];
} RECORD;
typedef struct _HEADER { /* File header descriptor */
	DWORD		NumRecords;
	DWORD		NumNonEmptyRecords;
} HEADER;

int _tmain (int argc, LPTSTR argv [])
{
	HANDLE hFile;
	LARGE_INTEGER CurPtr;
	DWORD FPos, OpenOption, nXfer, RecNo, FsFlags;
	RECORD Record;
	TCHAR String[STRING_SIZE], Command, Extra;
	OVERLAPPED ov = {0, 0, 0, 0, NULL}, ovZero = {0, 0, 0, 0, NULL};
	HEADER Header = {0, 0}; 
	SYSTEMTIME CurrentTime;
	BOOLEAN HeaderChange, RecordChange;

 	if (argc < 2)
		ReportError (_T ("Usage: RecordAccess file [nrec]"), 1, FALSE);
	OpenOption = (argc == 2) ? OPEN_EXISTING : CREATE_ALWAYS;
	hFile = CreateFile (argv [1], GENERIC_READ | GENERIC_WRITE,
		0, NULL, OpenOption, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
		ReportError (_T ("RecordAccess error: Cannot open existing file."), 2, TRUE);

	if (argc >= 3)  { /* Write the header (all header and pre-size the new file) */
#if 0
/* First, make the file sparse if possible - Recommended for large files */
/*      Challenge - get this to work properly - alternatively, make it sparse */
/*		adminstratively. This code, taken from several MS examples, fails on XP Home. */

		if (!GetVolumeInformation (_T("C:\\"), NULL, 0, NULL, NULL, &FsFlags , NULL, 0))
			ReportError (_T("RecordAccess Error: GetVolumeInformation."), 3, TRUE);
		if (FsFlags & FILE_SUPPORTS_SPARSE_FILES) {
			if ((FsFlags & FILE_SUPPORTS_SPARSE_FILES) &&
				!DeviceIoControl (hFile, FSCTL_SET_SPARSE, NULL, 0, NULL, 0, &nXfer, NULL)) 
				ReportError (_T("RecordAccess Error: Making new file sparse."), 0, TRUE);
		} 
		_tprintf (_T("Is it sparse? %x\n"), 
			GetFileAttributes(argv[1]) & FILE_ATTRIBUTE_SPARSE_FILE);
#endif
		Header.NumRecords = atoi(argv[2]);
		if (!WriteFile(hFile, &Header, sizeof (Header), &nXfer, &ovZero))
			ReportError (_T ("RecordAccess Error: WriteFile header."), 4, TRUE);
		CurPtr.QuadPart = sizeof(RECORD) * atoi(argv[2]) + sizeof(HEADER);
		FPos = SetFilePointer (hFile, CurPtr.LowPart, &CurPtr.HighPart, FILE_BEGIN);
		if (FPos == 0xFFFFFFFF && GetLastError () != NO_ERROR)
			ReportError (_T ("RecordAccess Error: Set Pointer."), 4, TRUE);
		if (!SetEndOfFile(hFile))
			ReportError (_T ("RecordAccess Error: Set End of File."), 5, TRUE);
	}

	/* Read the file header to find the number of records and non-empty records */
	if (!ReadFile(hFile, &Header, sizeof (HEADER), &nXfer, &ovZero))
		ReportError (_T ("RecordAccess Error: ReadFile header."), 6, TRUE);
	_tprintf (_T("File %s contains %d non-empty records of size %d. Total capacity is: %d\n"),
		argv[1], Header.NumNonEmptyRecords, sizeof(RECORD), Header.NumRecords);

	/* Prompt the user to read or write a numbered record */
	while (TRUE) {
		HeaderChange = FALSE; RecordChange = FALSE;
		_tprintf (_T("Enter r(ead)/w(rite)/d(elete)/q(uit) Record#\n"));
		_tscanf (_T ("%c" "%d" "%c"), &Command, &RecNo, &Extra );
		if (Command == 'q') break;
		if (RecNo >= Header.NumRecords) { 
			_tprintf (_T("Record Number is too large. Try again.\n"));
			continue;
		}
		CurPtr.QuadPart = RecNo * sizeof(RECORD) + sizeof(HEADER);
		ov.Offset = CurPtr.LowPart;
		ov.OffsetHigh = CurPtr.HighPart;
		if (!ReadFile (hFile, &Record, sizeof (RECORD), &nXfer, &ov))
			ReportError (_T ("RecordAccess: ReadFile failure."), 7, FALSE);
		GetSystemTime (&CurrentTime); /* Use to update record time fields */
		Record.RecordLastRefernceTime = CurrentTime;
		if (Command == 'r' || Command == 'd') { /* Report record contents, if any */
			if (Record.ReferenceCount == 0) {
				_tprintf (_T("Record Number %d is empty.\n"), RecNo);
				continue;
			} else {
				_tprintf (_T("Record Number %d. Reference Count: %d \n"), 
					RecNo, Record.ReferenceCount);
				_tprintf (_T("Data: %s\n"), Record.DataString);
				/* Exercise: Display times. See ls.c for an example */
				RecordChange = TRUE;
			}
			if (Command == 'd') { /* Delete the record */
				/* Challenge: In a sparse file, implement this with the
				   FSCTL_SET_ZERO_DATA flag on DeviceIoControl() */
				_tprintf (_T("About to delete this record.\n"));
				Record.ReferenceCount = 0;
				Header.NumNonEmptyRecords--;
				HeaderChange = TRUE;
				RecordChange = TRUE;
			}
		} else if (Command == 'w') { /* Write the record, even if for the first time */
			_tprintf (_T("Enter new data string for the record.\n"));
			_getts (String);
			if (Record.ReferenceCount == 0) {
				Record.RecordCreationTime = CurrentTime;
				Header.NumNonEmptyRecords++;
				HeaderChange = TRUE;
			}
			Record.RecordUpdateTime = CurrentTime;
			Record.ReferenceCount++;
			_tcsncpy (Record.DataString, String, STRING_SIZE-1);
			RecordChange = TRUE;
		} else {
			_tprintf (_T("Command must be r, w, or d. Try again.\n"));
		}
		/* Update the record in place if any record contents have changed. */
		if (RecordChange && !WriteFile (hFile, &Record, sizeof (RECORD), &nXfer, &ov))
			ReportError (_T ("RecordAccess: WriteFile update failure."), 8, FALSE);
		/* Update the number of non-empty records if required */
		if (HeaderChange) {
			if (!WriteFile (hFile, &Header, sizeof (Header), &nXfer, &ovZero))
				ReportError (_T ("RecordAccess: WriteFile update failure."), 9, FALSE);
		}
	}

	_tprintf (_T("Computed number of non-empty records is: %d\n"), Header.NumNonEmptyRecords);
	if (!ReadFile(hFile, &Header, sizeof (HEADER), &nXfer, &ovZero))
		ReportError (_T ("RecordAccess Error: ReadFile header."), 10, TRUE);
	_tprintf (_T("File %s NOW contains %d non-empty records. Total capacity is: %d\n"),
		argv[1], Header.NumNonEmptyRecords, Header.NumRecords);

	CloseHandle (hFile);
	return 0;
}
